/*
  Copyright 2018-2020, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2020, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  SPDX-License-Identifier: MIT
*/


#include <daemon/relocation/config_manager.hpp>
#include <daemon/daemon.hpp>
#include <global/global_defs.hpp>
#include <daemon/handler/rpc_defs.hpp>
#include <daemon/handler/rpc_util.hpp>
#include <global/hostsfile_util.hpp>
#include <global/hostsfile_util.hpp>
#include <global/rpc/rpc_types.hpp>

#include <cassert>
#include <iostream>

#include <fmt/format.h>

// margo
extern "C" {
#include <margo.h>
}

using namespace std;

namespace gkfs::relocation {

vector<pair<string, string>>
read_hosts_file() {
    string hostfile = GKFS_DATA->hosts_file();

    vector<pair<string, string>> hosts;
    try {
        hosts = gkfs::utils::load_hostfile(hostfile);
    } catch(const exception& e) {
        auto emsg = fmt::format("Failed to load hosts file: {}", e.what());
        throw runtime_error(emsg);
    }

    if(hosts.empty()) {
        throw runtime_error(fmt::format("Hostfile empty: '{}'", hostfile));
    }

    GKFS_DATA->spdlogger()->info("Hosts pool size: {}", hosts.size());

    return hosts;
}

/**
 * Send RPCs to all daemons in hosts file to trigger relocation.
 */
void
do_relocation() {
    auto hosts = read_hosts_file();

    cout << "number of hosts: " << hosts.size() << "\n";
    auto mid =
            margo_init(GKFS_DATA->bind_addr().c_str(), MARGO_CLIENT_MODE, 0, 0);
    if(mid == MARGO_INSTANCE_NULL) {
        throw runtime_error("Failed to initialize the Margo RPC client");
    }

    hg_id_t relocation_start_id = MARGO_REGISTER(
            mid, gkfs::rpc::tag::relocation_start, rpc_relocation_start_in_t,
            rpc_err_out_t, rpc_srv_relocation_start);


    uint64_t host_counter = -1;
    for(auto host : hosts) {
        host_counter++;
        cout << fmt::format("Hostname: {} Address: {} ID: {}\n", host.first,
                            host.second, host_counter);
        hg_addr_t host_addr;
        rpc_relocation_start_in_t in{};
        rpc_err_out_t out{};
        in.host_id = host_counter;
        auto ret = margo_addr_lookup(mid, host.second.c_str(), &host_addr);
        assert(ret == HG_SUCCESS);

        // let's do this sequential first
        hg_handle_t handle;
        ret = margo_create(mid, host_addr, relocation_start_id, &handle);
        assert(ret == HG_SUCCESS);

        ret = margo_forward(handle, &in); // blocking
        assert(ret == HG_SUCCESS);

        ret = margo_get_output(handle, &out);
        assert(ret == HG_SUCCESS);

        // TODO(dauer) process output

        if(HG_SUCCESS !=
           gkfs::rpc::margo_client_cleanup(&handle, &out, &mid, &host_addr)) {
            cout << "Error during margo cleanup.\n";
        }
    }
    margo_finalize(mid);
}

} // namespace gkfs::relocation